from enum import Enum


class OrderSide(Enum):
    SELL = 'SELL'
    BUY = 'BUY'

    @staticmethod
    def from_value(value):

        if isinstance(value, str):
            for order_side in OrderSide:

                if value.upper() == order_side.value:
                    return order_side
        elif isinstance(value, OrderSide):
            for order_side in OrderSide:

                if value == order_side:
                    return order_side

        raise ValueError("Could not convert value to OrderSide")

    def equals(self, other):

        if isinstance(other, Enum):
            return self.value == other.value

        else:
            return OrderSide.from_string(other) == self

    def __str__(self):
        return str(self.value)



from eltyer.models.order_status import OrderStatus
from eltyer.models.order_side import OrderSide
from eltyer.models.order_type import OrderType
from eltyer.models.model import Model


class Order(Model):

    def __init__(
        self,
        id,
        order_reference,
        side,
        type,
        target_symbol,
        trading_symbol,
        status,
        initial_price=None,
        amount_target_symbol=None,
        amount_trading_symbol=None,
    ):
        self.id = id
        self.order_reference = order_reference
        self.side = OrderSide.from_value(side)
        self.type = OrderType.from_value(type)
        self.target_symbol = target_symbol
        self.trading_symbol = trading_symbol
        self.status = OrderStatus.from_value(status)
        self.initial_price = initial_price
        self.amount_trading_symbol = amount_trading_symbol
        self.amount_target_symbol = amount_target_symbol

    @staticmethod
    def from_dict(data):
        return Order(
            id=data.get("id"),
            order_reference=data.get("order_reference"),
            side=data.get("side"),
            type=data.get("type"),
            target_symbol=data.get("target_symbol"),
            trading_symbol=data.get("trading_symbol"),
            status=data.get("status"),
            initial_price=data.get("initial_price"),
            amount_target_symbol=data.get("amount_target_symbol"),
            amount_trading_symbol=data.get("amount_trading_symbol"),
        )

    def __repr__(self):
        return self.repr(
            id=self.id,
            order_reference=self.order_reference,
            side=str(self.side),
            type=str(self.type),
            target_symbol=self.target_symbol,
            trading_symbol=self.trading_symbol,
            status=str(self.status),
            initial_price=self.initial_price,
            amount_target_symbol=self.amount_target_symbol,
            amount_trading_symbol=self.amount_trading_symbol,
        )



import os
from eltyer.configuration.constants import BASE_URL, ELTYER_API_KEY, API_KEY


class Config(dict):
    LOG_LEVEL = 'INFO'
    APP_DIR = os.path.abspath(os.path.dirname(__file__))
    PROJECT_ROOT = os.path.abspath(os.path.join(APP_DIR, os.pardir))

    # ENDPOINTS
    ORDERS_ENDPOINT = "/order-service/v1/orders"
    LIST_ORDERS_ENDPOINT = "/order-service/v1/algorithms/{algorithm_id}/orders"
    API_KEY_VERIFY_ENDPOINT = "/order-service/v1/algorithms/api-keys/verify"
    POSITIONS_ENDPOINT = "/order-service/v1/algorithms/" \
                         "{algorithm_id}/positions"
    PORTFOLIO_ENDPOINT = "/order-service/v1/algorithms/" \
                         "{algorithm_id}/portfolio"

    API_KEY = None
    HOST_ORDER_SERVICE = BASE_URL

    def __init__(self):
        super().__init__()

        for attribute_key in dir(Config):

            if attribute_key.isupper():
                self[attribute_key] = getattr(self, attribute_key)

    def __setitem__(self, key, item):
        self.__dict__[key] = item

    def __getitem__(self, key):
        return self.__dict__[key]

    def __repr__(self):
        return repr(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def __delitem__(self, key):
        del self.__dict__[key]

    def clear(self):
        return self.__dict__.clear()

    def copy(self):
        return self.__dict__.copy()

    def has_key(self, k):
        return k in self.__dict__

    def update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def pop(self, *args):
        return self.__dict__.pop(*args)

    def __cmp__(self, dict_):
        return self.__cmp__(self.__dict__, dict_)

    def __contains__(self, item):
        return item in self.__dict__

    def __iter__(self):
        return iter(self.__dict__)

    def __str__(self):
        field_strings = []

        for attribute_key in self:

            if attribute_key.isupper():
                field_strings.append(
                    f'{attribute_key}='
                    f'{self[attribute_key]!r}'
                )

        return f"<{self.__class__.__name__}({','.join(field_strings)})>"

    def get(self, key: str, default=None):
        """
        Mimics the dict get() functionality
        """

        try:
            return self[key]
        # Ignore exception
        except Exception:
            pass

        return default

    def set(self, key: str, value) -> None:
        self[key] = value

    def from_dict(self, dictionary):

        for attribute_key in dictionary:
            if attribute_key:
                self.set(attribute_key, dictionary[attribute_key])
                self[attribute_key] = dictionary[attribute_key]

    def from_env(self):
        api_key = os.getenv(ELTYER_API_KEY, None)
        self.set(API_KEY, api_key)
        self[API_KEY] = api_key

    @property
    def configured(self):
        return not (self.API_KEY is None or self.HOST_ORDER_SERVICE is None)



from eltyer.models.model import Model


class Position(Model):

    def __init__(self, id, symbol, amount, percentage, orders):
        self.id = id
        self.symbol = symbol
        self.amount = amount
        self.percentage = percentage
        self.orders = orders

    @staticmethod
    def from_dict(data):
        return Position(
            id=data.get("id", None),
            symbol=data.get("symbol", None),
            amount=data.get("amount", None),
            percentage=data.get("percentage", None),
            orders=data.get("orders", None)
        )

    def __repr__(self):
        return self.repr(
            id=self.id,
            symbol=self.symbol,
            amount=self.amount,
            percentage=self.percentage,
            orders=self.orders
        )



from enum import Enum


class OrderStatus(Enum):
    SUCCESS = 'SUCCESS'
    PENDING = 'PENDING'
    TO_BE_SENT = "TO_BE_SENT"
    FAILED = "FAILED"
    CANCELED = "CANCELED"
    CLOSED = "CLOSED"

    @staticmethod
    def from_value(value):

        if isinstance(value, str):
            for order_type in OrderStatus:

                if value.upper() == order_type.value:
                    return order_type
        elif isinstance(value, OrderStatus):
            for status in OrderStatus:

                if value == status:
                    return status

        return None

    def equals(self, other):

        if other is None:
            return False

        if isinstance(other, Enum):
            return self.value == other.value

        else:
            return OrderStatus.from_string(other) == self

    def __str__(self):
        return str(self.value)



from abc import abstractmethod


class Model:

    @staticmethod
    @abstractmethod
    def from_dict(data):
        pass

    def repr(self, **fields) -> str:
        """
        Helper for __repr__
        """

        field_strings = []

        for key, field in fields.items():
            field_strings.append(f'{key}={field!r}')

        return f"<{self.__class__.__name__}({','.join(field_strings)})>"



from enum import Enum


class OrderType(Enum):
    LIMIT = 'LIMIT'
    MARKET = 'MARKET'

    @staticmethod
    def from_value(value):

        if isinstance(value, str):
            for order_type in OrderType:

                if value.upper() == order_type.value:
                    return order_type
        elif isinstance(value, OrderType):
            for order_type in OrderType:

                if value == order_type:
                    return order_type

        raise ValueError("Could not convert value to OrderType")

    def equals(self, other):

        if isinstance(other, Enum):
            return self.value == other.value

        else:
            return OrderType.from_value(other) == self

    def __str__(self):
        return str(self.value)



from eltyer.models.model import Model


class Portfolio(Model):

    def __init__(
        self,
        id,
        algorithm_id,
        trading_symbol,
        delta,
        allocated,
        allocated_percentage,
        unallocated,
        unallocated_percentage,
        broker,
        realized,
        created_at,
        updated_at,
        orders,
        positions,
        **kwargs
    ):
        self.id = id
        self.algorithm_id = algorithm_id
        self.delta = delta
        self.allocated = allocated
        self.allocated_percentage = allocated_percentage
        self.trading_symbol = trading_symbol
        self.unallocated = unallocated
        self.unallocated_percentage = unallocated_percentage
        self.realized = realized
        self.orders = orders
        self.positions = positions
        self.broker = broker
        self.created_at = created_at
        self.updated_at = updated_at
        super(Portfolio, self).__init__(**kwargs)

    @staticmethod
    def from_dict(data):
        return Portfolio(
            id=data.get("id", None),
            algorithm_id=data.get("algorith_id", None),
            delta=data.get("delta", None),
            allocated=data.get("allocated", None),
            allocated_percentage=data.get("allocated_percentage", None),
            trading_symbol=data.get("trading_symbol", None),
            unallocated=data.get("unallocated", None),
            unallocated_percentage=data.get("unallocated_percentage", None),
            realized=data.get("realized", None),
            orders=data.get("orders", None),
            positions=data.get("positions", None),
            broker=data.get("broker", None),
            created_at=data.get("created_at", None),
            updated_at=data.get("updated_at", None),
        )

    def __repr__(self):
        return self.repr(
            id=self.id,
            algorithm_id=self.algorithm_id,
            delta=self.delta,
            allocated=self.allocated,
            allocated_percentage=self.allocated_percentage,
            trading_symbol=self.trading_symbol,
            unallocated=self.unallocated,
            unallocated_percentage=self.unallocated_percentage,
            realized=self.realized,
            orders=self.orders,
            positions=self.positions,
            broker=self.broker,
            created_at=self.created_at,
            updated_at=self.updated_at,
        )



def get_version(version=None):
    version = get_complete_version(version)
    main = get_main_version(version)
    return main


def get_main_version(version=None):
    """Return main version (X.Y[.Z]) from VERSION."""
    version = get_complete_version(version)
    parts = 2 if version[2] == 0 else 3
    return '.'.join(str(x) for x in version[:parts])


def get_complete_version(version=None):
    """
    Return a tuple of the investing algorithm framework  version. If version
    argument is non-empty,
    check for correctness of the tuple provided.
    """
    if version is None:
        from eltyer import VERSION as version
    else:
        assert len(version) == 5
        assert version[3] in ('alpha', 'beta', 'rc', 'final')

    return version



import setuptools
from eltyer import get_version

VERSION = get_version()

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    required = f.read().splitlines()

setuptools.setup(
    name="eltyer",
    version=get_version(),
    license='Apache License 2.0',
    author="ELTYER",
    description="Official python client for ELTYER",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ELTYER/eltyer-python-client.git",
    packages=setuptools.find_packages(exclude=['tests', 'tests.*']),
    keywords=[
        'ELTYER',
        'eltyer',
        'investing-algorithm',
        'INVESTING',
        'BOT',
        'ALGORITHM',
        'FRAMEWORK',
        'investing-bots',
        'trading-bots'
    ],
    classifiers=[
        "Intended Audience :: Developers",
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        "Topic :: Software Development",
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
    ],
    install_requires=required,
    python_requires='>=3',
    include_package_data=True,
)



from eltyer.models.order_type import OrderType
from eltyer.models.order_side import OrderSide
from eltyer.models.order_status import OrderStatus
from eltyer.models.order import Order
from eltyer.models.position import Position
from eltyer.models.portfolio import Portfolio

__all__ = [
    "OrderType",
    "OrderSide",
    "OrderStatus",
    "Order",
    "Position",
    "Portfolio"
]



import requests
from logging import getLogger
from multiprocessing.pool import ThreadPool

from eltyer.configuration.config import Config
from eltyer.utils.version import get_version
from eltyer.models import OrderSide, OrderType, Order, Position, Portfolio, \
    OrderStatus
from eltyer.exceptions import ClientException

VERSION = (0, 1, 1, 'alpha', 0)

logger = getLogger(__name__)


class Client:
    config = Config()
    thread_count = 2
    _pool = None

    # Algorithm specific attributes
    algorithm_id = None
    environment = None

    def start(self):

        if not self.config.configured:
            raise ClientException("Client is not configured")

        algorithm_data = self._retrieve_algorithm()
        Client.algorithm_id = algorithm_data["algorithm_id"]
        Client.environment = algorithm_data["environment"]

    def stop(self):

        if self._pool:
            self._pool.close()
            self._pool.join()
            self._pool = None

    def get_environment(self):
        return self.environment

    @property
    def pool(self):
        if self._pool is None:
            self._pool = ThreadPool(self.thread_count)

        return self._pool

    def create_limit_order(
        self,
        target_symbol: str,
        price: float,
        amount: float,
        side: str = OrderSide.BUY.value
    ) -> Order:
        self.check_context()

        payload = {
            "target_symbol": target_symbol,
            "price": price,
            "amount": amount,
            "side": OrderSide.from_value(side).value,
            "type": OrderType.LIMIT.value,
        }

        response = requests.post(
            f"{self.config.HOST_ORDER_SERVICE}{self.config.ORDERS_ENDPOINT}",
            json=payload,
            headers={"x-api-key": self.config.API_KEY}
        )

        data = self._handle_response(response)
        return Order.from_dict(data)

    def create_market_order(
        self,
        target_symbol: str,
        amount: float,
    ) -> Order:
        payload = {
            "target_symbol": target_symbol,
            "amount": amount,
            "side": OrderSide.SELL.value,
            "type": OrderType.MARKET.value,
        }

        response = requests.post(
            f"{self.config.HOST_ORDER_SERVICE}{self.config.ORDERS_ENDPOINT}",
            json=payload,
            headers={"x-api-key": self.config.API_KEY}
        )

        data = self._handle_response(response)
        return Order.from_dict(data)

    def get_orders(self, target_symbol: str = None, status=None):
        self.check_context()

        params = {}

        if target_symbol is not None:
            params["target_symbol"] = target_symbol

        if status is not None:
            params["status"] = OrderStatus.from_value(status)

        params["itemized"] = True

        response = requests.get(
            f"{self.config.HOST_ORDER_SERVICE}"
            f"{self.config.LIST_ORDERS_ENDPOINT.format(algorithm_id=self.algorithm_id)}",
            params=params,
            headers={"x-api-key": self.config.API_KEY}
        )

        data = self._handle_response(response)
        orders = []

        for order_data in data["items"]:
            orders.append(Order.from_dict(order_data))

        return orders

    def get_order(self, reference_id) -> Order:
        response = requests.get(
            f"{self.config.HOST_ORDER_SERVICE}"
            f"{self.config.LIST_ORDERS_ENDPOINT.format(algorithm_id=self.algorithm_id)}",
            headers={"x-api-key": self.config.API_KEY}
        )

        data = self._handle_response(response)

        for order_data in data["items"]:
            ref_order = Order.from_dict(order_data)

            if ref_order.order_reference == reference_id:
                return ref_order

        return None

    def get_positions(self):
        self.check_context()

        response = requests.get(
            f"{self.config.HOST_ORDER_SERVICE}"
            f"{self.config.POSITIONS_ENDPOINT.format(algorithm_id=self.algorithm_id)}",
            params={"itemized": True},
            headers={"x-api-key": self.config.API_KEY}
        )

        data = self._handle_response(response)
        positions = []

        for position_data in data["items"]:
            positions.append(Position.from_dict(position_data))

        return positions

    def get_position(self, symbol: str):
        self.check_context()

        response = requests.get(
            f"{self.config.HOST_ORDER_SERVICE}"
            f"{self.config.POSITIONS_ENDPOINT.format(algorithm_id=self.algorithm_id)}",
            params={"itemized": True},
            headers={"x-api-key": self.config.API_KEY}
        )

        data = self._handle_response(response)

        for position_data in data["items"]:

            if position_data["symbol"].upper() == symbol.upper():
                return Position.from_dict(position_data)

        return None

    def get_portfolio(self):
        self.check_context()

        response = requests.get(
            f"{self.config.HOST_ORDER_SERVICE}"
            f"{self.config.PORTFOLIO_ENDPOINT.format(algorithm_id=self.algorithm_id)}",
            params={"itemized": True},
            headers={"x-api-key": self.config.API_KEY}
        )

        return Portfolio.from_dict(self._handle_response(response))

    def _handle_response(self, response):

        if response.status_code >= 400:

            if response.status_code == 500:
                raise ClientException("Something went wrong at ELTYER")

            raise ClientException(response.json()["error_message"])

        return response.json()

    def _retrieve_algorithm(self):
        response = requests.get(
            f"{self.config.HOST_ORDER_SERVICE}"
            f"{self.config.API_KEY_VERIFY_ENDPOINT}",
            headers={"x-api-key": self.config.API_KEY}
        )

        return self._handle_response(response)

    def check_context(self):

        if self.algorithm_id is None:
            raise ClientException("Client is not configured")


__all__ = [
    "Client",
    "get_version",
    "OrderType",
    "OrderStatus",
    "OrderSide",
    "ClientException"
]
